import math

import torch
from torch.utils.data import DataLoader
import torch.optim as optim
from typing import List
from torch import Tensor
from .mortm import MORTM
from .tokenizer import Tokenizer, PITCH_TYPE

#from .train import _set_train_data
from .progress import _DefaultLearningProgress, LearningProgress
from .datasets import MORTM_DataSets
from .aya_node import Token, _get_symbol
import os


def remove_subsequence_tensor(a, b):
    len_b = b.shape[0]
    for i in range(a.shape[0] - len_b + 1):
        if torch.equal(a[i:i+len_b], b):
            return torch.cat((a[:i], a[i+len_b:]))
    return a


def get_convert_measure_list(sequence):
    '''
    シーケンスから小節の区切り目のトークンを境目にスプリットし、その配列を返します。
    :param sequence:  小節ごとに区切られたシーケンスの配列
    :return: sequence
    '''
    segments = []
    start_idx = None
    for i, val in enumerate(sequence):
        #print(i, val)
        if val == 3:
            if start_idx is not None and i > start_idx:
                segments.append(sequence[start_idx:i])
            start_idx = i + 1
    return segments


def compose_sequence_reward(sequence: Tensor, tokenizer: Tokenizer):
    '''
    シーケンスを受け取り、正しい生成の順番であるかを評価します。
    通常は S -> P -> Dで生成されるはずです.


    :param sequence: 一小節分のシーケンス
    :param tokenizer: MORTMのトークナイザー
    :return: 報酬
    '''

    reward_count = 0
    token_list: List[Token] = tokenizer.token_list[1:]
    token_count = 0
    for seq in enumerate(sequence):
        if not token_list[token_count].is_my_token(seq):
            reward_count += 1

        token_count += 1
        if len(token_list) <= token_count:
            token_count = 0

    reward = _calc_loss(reward_count if token_count == 0 else 10, 0, 0.01)

    return reward




def compose_measure_reward(measure, tokenizer: Tokenizer, last_duration):
    pitch_token: Token = tokenizer.get_token_converter(PITCH_TYPE)

    measure = measure[(pitch_token.start <= measure) | (measure >= pitch_token.end)]
    i = 0
    duration = 0
    ld = 0
    while i < len(measure):
        if i == 0:
            if last_duration is None:
                duration += _get_symbol(tokenizer.rev_get(measure[i].item()))
            else:
                duration += _get_symbol(tokenizer.rev_get(measure[i].item())) - last_duration
        elif i == len(measure) - 1:
            ld = _get_symbol(tokenizer.rev_get(measure[i].item()))
            if duration < 64 < duration + ld:
                duration = 64
            else:
                duration += ld
        elif i % 2 == 0:
            duration += _get_symbol(tokenizer.rev_get(measure[i].item())) - _get_symbol(tokenizer.rev_get(measure[i - 1].item()))
        else:
            duration += _get_symbol(tokenizer.rev_get(measure[i].item()))

        i += 1
    reward = _calc_loss(duration, 64, 0.001)
    return reward, ld


def _calc_loss(x, n:int, k:float):
    return math.e ** k * abs(x - n) - 1


def _reward_function(base_seq, sequence, tokenizer: Tokenizer):
    if base_seq is not None:
        seq = remove_subsequence_tensor(sequence, base_seq)
    else:
        seq = sequence
    measure_list = get_convert_measure_list(seq)
    reward = 100 if len(measure_list) is 0 else 0
    last_duration = None
    for measure in measure_list:
        cmr, ld = compose_measure_reward(measure, tokenizer, last_duration)
        cmr /= len(measure_list)
        last_duration = ld
        csr = compose_sequence_reward(measure, tokenizer) / len(measure_list)
        reward += (csr + cmr)

    return reward


def calc_mortm_loss(probs: Tensor, reward):
    loss = -(reward * probs.sum()) / 200
    return loss


def re_train(epoch, model: MORTM, dataset_directory:str, tokenizer: Tokenizer,
             lr=8e-6, progress:LearningProgress =_DefaultLearningProgress(), goal_loss= 1.0):
    direct = os.listdir(dataset_directory)
    optimizer = optim.Adam(model.parameters(), lr=lr)
    dataset:MORTM_DataSets = _set_train_data(dataset_directory, direct, progress, is_fine_turing=False)
    loader = DataLoader(dataset, batch_size=1, shuffle=True)

    for e in range(1, epoch + 1):
        model.train()
        optimizer.zero_grad()

        for seq in loader:
            is_reinforcement_goal = False
            reinforce_counter = 0
            while not is_reinforcement_goal:
                input_seq: Tensor = seq.clone().to(progress.get_device())
                input_seq = input_seq.squeeze(0)
                input_seq = input_seq[:-1]
                model.eval()
                # サンプル
                #sample, log_probs = model.top_k_sampling_length_encoder(seq, temperature=0.9, top_k=8, max_length=500)

                # ベースライン(argmax)
                baseline, log_probs = model.argmax_sampling_encoder(input_seq, max_length=200)

 #               sample_reward = _reward_function(seq, sample, tokenizer)
                baseline_reward = _reward_function(input_seq, baseline, tokenizer)


                model.train()
                #loss = calculate_scst_loss(log_probs, sample_reward, baseline_reward)
                loss = calc_mortm_loss(log_probs, baseline_reward)
                print(f"現在の損失は[{loss: 4f}]になっています。")

                loss.backward()
#                for param in model.parameters():
#                     print(param.grad)  # 勾配がNoneなら問題あり
                torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.5)
                print("更新中...")
                optimizer.step()
                print("リセット中....")
                optimizer.zero_grad()
                if reinforce_counter % 1000 == 0:
                    torch.save(model.state_dict(), f"./out/model/MORTM.re.{epoch}.{reinforce_counter}_loss.{loss:.4f}.pth")

                if loss < goal_loss:
                    is_reinforcement_goal = True
                    print("次のStepに移動します。")
                    torch.save(model.state_dict(), f"./out/model/MORTM.re.{epoch}.{reinforce_counter}_loss.{loss:.4f}.pth")

                del loss  # 損失を削除
                torch.cuda.empty_cache()
                reinforce_counter += 1

            pass



import torch
from torch import Tensor
import torch.nn as nn

from .PositionalEncoding import PositionalEncoding
from .RelativePositionalRepresentations import CustomTransformerEncoder
from .rpr import TransformerEncoderLayerRPR, TransformerEncoderRPR
from .progress import LearningProgress
from torch.distributions import Categorical

class MORTM(nn.Module):
    token_dict = {
        'SHIFT': [1, 6],
        'START': [7, 38],
        'PITCH': [39, 166],
        'DURATION': [167, 266]
    }

    def __init__(self, vocab_size, progress: LearningProgress, trans_layer=6, num_heads=8, d_model=512,
                 dim_feedforward=1024, dropout=0.1,
                 position_length=2048, use_rpr=True, use_decoder=False):
        super(MORTM, self).__init__()

        if use_decoder:
            print("Use decoder mode!")

        self.progress = progress
        self.trans_layer = trans_layer
        self.num_heads = num_heads
        self.d_model = d_model
        self.dim_feedforward = dim_feedforward
        self.dropout = dropout

        #位置エンコーディングを作成
        #self.positional: LearnablePositionalEncoding = LearnablePositionalEncoding(self.d_model, progress, dropout, position_length).to(self.progress.get_device())
        self.positional: PositionalEncoding = PositionalEncoding(self.d_model, progress, dropout, position_length).to(self.progress.get_device())
        #Transformerの設定
        if not use_rpr:
            self.transformer: nn.Transformer = nn.Transformer(d_model=self.d_model, nhead=num_heads,  #各種パラメーターの設計
                                                              num_encoder_layers=self.trans_layer,
                                                              num_decoder_layers= self.trans_layer if use_decoder else 0, #デコーダーは使うときと使わない時がある。
                                                              dropout=self.dropout, dim_feedforward=dim_feedforward,
                                                              custom_decoder=None if use_decoder else DummyDecoder()
                                                              ).to(self.progress.get_device())
        else:
            encoder_norm = nn.LayerNorm(self.d_model)
            encoder_layer = TransformerEncoderLayerRPR(self.d_model, self.num_heads, self.dim_feedforward, self.dropout, er_len=position_length)
            encoder = TransformerEncoderRPR(encoder_layer, self.trans_layer, encoder_norm)
            self.transformer = nn.Transformer(
                d_model=self.d_model, nhead=self.num_heads, num_encoder_layers=self.trans_layer,
                num_decoder_layers=self.trans_layer if use_decoder else 0,
                dropout=self.dropout, # activation=self.ff_activ,
                dim_feedforward=self.dim_feedforward,
                custom_decoder=None if use_decoder else DummyDecoder(),
                custom_encoder=encoder
            ).to(device=progress.get_device())

            print("Use RPR Transformer")
        print(f"Input Vocab Size:{vocab_size}")
        self.Wout: nn.Linear = nn.Linear(self.d_model, vocab_size).to(self.progress.get_device())

        self.embedding: nn.Embedding = nn.Embedding(vocab_size, self.d_model).to(self.progress.get_device())
        self.softmax: nn.Softmax = nn.Softmax(dim=-1).to(self.progress.get_device())

    def forward(self, inputs_seq, tgt_seq=None, src_mask=None, tgt_mask=None, input_padding_mask=None, tgt_padding_mask=None):
        if tgt_seq is None or src_mask is None:
            mask = self.transformer.generate_square_subsequent_mask(inputs_seq.shape[1]).to(self.progress.get_device())
        elif src_mask is not None:
            mask = src_mask
        else:
            mask = None

        inputs_em: Tensor = self.embedding(inputs_seq)
        inputs_em = inputs_em.permute(1, 0, 2)

        inputs_pos: Tensor = self.positional(inputs_em)

        if tgt_seq is not None:
            tgt_e = self.embedding(tgt_seq)
            tgt_e = tgt_e.permute(1, 0, 2)
            tgt_p = self.positional(tgt_e)
        else:
            tgt_p = inputs_pos

        out: Tensor = self.transformer(inputs_pos, tgt_p, src_mask=mask, tgt_mask=tgt_mask,
                                       src_key_padding_mask=input_padding_mask, tgt_key_padding_mask=tgt_padding_mask)

        out.permute(1, 0, 2)

        score:Tensor = self.Wout(out)
        return score.to(self.progress.get_device())

    def top_k_sampling_length_encoder(self, input_sequence, temperature=1.0, top_k=3, max_length=100):
        self.eval()
        """
        MORTMモデルにトップKサンプリングと温度シーケンスを実装する関数

        Args:
            input_sequence: 1次元の入力シーケンス (List[int] or torch.Tensor)。
            temperature: 温度パラメータ。デフォルトは1.0。
            top_k: サンプリングする上位K個のトークンの数。デフォルトは3。
            max_length: 生成するシーケンスの最大長さ。デフォルトは100。

        Returns:
            1次元の生成されたシーケンス (torch.Tensor)。
        """
        log_prob_list = []
        # 入力シーケンスをtorch.tensorに変換
        if not isinstance(input_sequence, torch.Tensor):
            input_sequence = torch.tensor(input_sequence, dtype=torch.long, device=self.progress.get_device())

        # 入力シーケンスの長さを取得
        input_length = input_sequence.size(0)

        # 生成されたシーケンスを格納するリスト
        generated_sequence = input_sequence.tolist()

        # 生成をループ
        for i in range(max_length):
            # モデルに渡すための入力の準備 (2次元に変換)
            input_tensor = input_sequence.unsqueeze(0)  # (1, sequence_length)
            #print(f"I{i} input_tensor")
            print(f"\r Generating...{i / max_length * 100}%", end="")
            # モデルに入力して次のトークンのスコアを取得 (3次元で返ってくる)
            with torch.no_grad():
                mask = self.transformer.generate_square_subsequent_mask(input_tensor.shape[1]).to(
                    self.progress.get_device())
                #print(input_tensor.shape)
                scores = self(input_tensor)  # (1, sequence_length, vocab_size)
                #print(f"SCORE: {scores.shape}")

            # 最新のトークンのスコアを取得 (最後のトークンに対するスコア)
            logits = scores[:, -1, :]  # (1, vocab_size)

            # 温度の適用
            logits = logits / temperature
            logits = logits[-1, :]
            # ソフトマックスを適用して確率を取得
            probs = self.softmax(logits)  # (vocab_size)

            # トップKの確率でトークンをフィルタリング
            topk_probs, topk_indices = torch.topk(probs, top_k)
            #print(sorted_probs, sorted_indices)

            # 再度正規化
            topk_probs = topk_probs / topk_probs.sum(dim=-1, keepdim=True)

            # トークンをサンプリング
            distribution = Categorical(topk_probs)
            sampled_index = distribution.sample()

            # ソートされたインデックスから元のインデックスに変換
            next_token = topk_indices[sampled_index].item()

            log_probs = torch.log(topk_probs[sampled_index])
            log_prob_list.append(log_probs)
#            print(next_token)

            # シーケンスにトークンを追加
            generated_sequence.append(next_token)

            # 次のステップの入力として準備
            input_sequence = torch.tensor(generated_sequence, dtype=torch.long, device=self.progress.get_device())

        return input_sequence, torch.tensor(log_prob_list, device=self.progress.get_device())

    def argmax_sampling_encoder(self, input_sequence, max_length=500):
        #self.eval()
        """
        MORTMモデルにトップKサンプリングと温度シーケンスを実装する関数

        Args:
            input_sequence: 1次元の入力シーケンス (List[int] or torch.Tensor)。
            temperature: 温度パラメータ。デフォルトは1.0。
            top_k: サンプリングする上位K個のトークンの数。デフォルトは3。
            max_length: 生成するシーケンスの最大長さ。デフォルトは100。

        Returns:
            1次元の生成されたシーケンス (torch.Tensor)。
        """
        log_prob_list = []
        # 入力シーケンスをtorch.tensorに変換
        if not isinstance(input_sequence, torch.Tensor):
            input_sequence = torch.tensor(input_sequence, dtype=torch.long, device=self.progress.get_device())

        # 入力シーケンスの長さを取得
        input_length = input_sequence.size(0)

        # 生成されたシーケンスを格納するリスト
        generated_sequence = input_sequence.tolist()

        # 生成をループ
        for i in range(max_length):
            # モデルに渡すための入力の準備 (2次元に変換)
            input_tensor = input_sequence.unsqueeze(0)  # (1, sequence_length)

            print(f"\r Generating...{i / max_length * 100}%", end="")
            # モデルに入力して次のトークンのスコアを取得 (3次元で返ってくる)
            mask = self.transformer.generate_square_subsequent_mask(input_tensor.shape[1]).to(
                self.progress.get_device())

            scores = self(input_tensor)  # (1, sequence_length, vocab_size)

            # 最新のトークンのスコアを取得 (最後のトークンに対するスコア)
            logits = scores[:, -1, :]  # (1, vocab_size)
            logits = logits[-1, :]


            # ソフトマックスを適用して確率を取得
            probs: Tensor = self.softmax(logits)  # (vocab_size)
            d = Categorical(probs)
            next_token = d.sample()

            log_probs = d.log_prob(next_token)
            log_prob_list.append(log_probs)
            # シーケンスにトークンを追加
            generated_sequence.append(next_token)

            # 次のステップの入力として準備
            input_sequence = torch.tensor(generated_sequence, dtype=torch.long, device=self.progress.get_device())

        return input_sequence, torch.stack(log_prob_list).to(self.progress.get_device())


    def top_k_sampling_length_decoder(self, input_sequence, temperature=1.0, top_k=3, max_length=100):
        self.eval()
        """
        MORTMモデルにトップKサンプリングと温度シーケンスを実装する関数

        Args:
            input_sequence: 1次元の入力シーケンス (List[int] or torch.Tensor)。
            temperature: 温度パラメータ。デフォルトは1.0。
            top_k: サンプリングする上位K個のトークンの数。デフォルトは3。
            max_length: 生成するシーケンスの最大長さ。デフォルトは100。

        Returns:
            1次元の生成されたシーケンス (torch.Tensor)。
        """

        # 入力シーケンスをtorch.tensorに変換
        if not isinstance(input_sequence, torch.Tensor):
            input_sequence = torch.tensor(input_sequence, dtype=torch.long, device=self.progress.get_device())

        # 入力シーケンスの長さを取得
        input_length = input_sequence.size(0)

        # 生成されたシーケンスを格納するリスト
        generated_sequence = [11]

        # 生成をループ
        for i in range(max_length):
            # モデルに渡すための入力の準備 (2次元に変換)
            tgt = torch.tensor(generated_sequence, device=self.progress.get_device()).unsqueeze(0)  # (1, sequence_length)
            src = input_sequence.unsqueeze(0)
            #print(f"I{i} input_tensor")

            # モデルに入力して次のトークンのスコアを取得 (3次元で返ってくる)
            with torch.no_grad():
                mask = self.transformer.generate_square_subsequent_mask(tgt.shape[1]).to(
                    self.progress.get_device())
                print(src.shape)
                scores = self(src, tgt_seq=tgt)  # (1, sequence_length, vocab_size)
                print(f"SCORE: {scores.shape}")

            # 最新のトークンのスコアを取得 (最後のトークンに対するスコア)
            logits = scores[:, -1, :]  # (1, vocab_size)

            # 温度の適用
            logits = logits / temperature
            logits = logits[-1, :]

            # ソフトマックスを適用して確率を取得
            probs = self.softmax(logits)  # (vocab_size)

            # トップKの確率でトークンをフィルタリング
            sorted_probs, sorted_indices = torch.topk(probs, top_k)
            #print(sorted_probs, sorted_indices)

            # 再度正規化
            sorted_probs = sorted_probs / sorted_probs.sum(dim=-1, keepdim=True)

            # トークンをサンプリング
            sampled_index = torch.multinomial(sorted_probs, 1).item()  # サンプリングされたインデックスを取得

            # ソートされたインデックスから元のインデックスに変換
            next_token = sorted_indices[sampled_index].item()

            #            print(next_token)

            # シーケンスにトークンを追加
            generated_sequence.append(next_token)

            # 次のステップの入力として準備
#            input_sequence = torch.tensor(generated_sequence, dtype=torch.long, device=self.progress.get_device())
        input_sequence = torch.cat((input_sequence, torch.tensor(generated_sequence, device=self.progress.get_device())))
        return input_sequence

    def top_p_sampling_length(self, input_seq, p=0.8, max_length=20, temperature=1.0):
        self.eval()
        if not isinstance(input_seq, torch.Tensor):
            input_seq = torch.tensor(input_seq, dtype=torch.long, device=self.progress.get_device())

        generated = input_seq.tolist()
        for i in range(max_length):
 #           print(f"INPUTS:   {input_seq}")

            input_seq = input_seq.unsqueeze(0)
            logits= self(input_seq)
            logits = logits[:, -1, :][-1, :]
            token = self.top_p_sampling(logits, p=p, temperature=temperature)
            generated.append(token)
            input_seq = torch.tensor(generated, dtype=torch.long, device=self.progress.get_device())
            if token == 2:
                print("終了宣言されたため、処理を中断します。")
                break
            print(f"\r Generating... {i / max_length}%", end="")

        return input_seq

    def top_p_sampling(self, logits, p=0.9, temperature=1.0)-> int:

        logits = logits / temperature
        # logitsをソフトマックスで確率分布に変換
        probs = self.softmax(logits)
        # 確率の降順に並べ替え、そのインデックスを取得
        sorted_probs, sorted_indices = torch.sort(probs, descending=True)

        # 累積確率を計算
        cumulative_probs = torch.cumsum(sorted_probs, dim=-1)

        # 累積確率がpを超えるインデックスを取得
        cutoff_index = torch.where(cumulative_probs > p)[0][0]

        # 上位pに入らないトークンの確率を0にする
        sorted_probs[cutoff_index + 1:] = 0

        # 確率を再正規化
        sorted_probs /= torch.sum(sorted_probs)

        # トークンをサンプリング
        sampled_index = torch.multinomial(sorted_probs, 1)

        # インデックスを元の順序に戻す
        return sorted_indices[sampled_index].item()


class DummyDecoder(nn.Module):
    def __init__(self):
        super(DummyDecoder, self).__init__()

    def forward(self, tgt, memory, tgt_mask, memory_mask, tgt_key_padding_mask, memory_key_padding_mask, **kwargs):
        return memory



import math


from torch.nn import CrossEntropyLoss, Softmax
from torch import Tensor
from typing import Optional
from .reinforcement import  _reward_function
from .tokenizer import Tokenizer


class ReinforceCrossEntropy(CrossEntropyLoss):

    def __init__(self, tokenizer: Tokenizer, k=1, warmup=100, weight: Optional[Tensor] = None, size_average=None,
                 ignore_index: int = -100,
                 reduce=None, reduction: str = 'mean', label_smoothing: float = 0.0) -> None:
        super().__init__(weight=weight, size_average=size_average, ignore_index=ignore_index, reduce=reduce,
                         reduction=reduction, label_smoothing=label_smoothing)
        self.softmax = Softmax(dim=1)
        self.tokenizer = tokenizer
        self.cs: float = 1
        self.te: float = 0
        self.stepup = 0
        self.k = k
        self.warmup = warmup

        #self.step()

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        score: Tensor = self.softmax(input)
        seq = score.argmax(dim=1)
        l = _reward_function(None, seq, self.tokenizer)

        return self.cs * super().forward(input, target) + self.te * l

    def step(self):

        self.stepup += 1
        self.te =0.5 * (1 + math.cos(math.pi * (self.stepup - (2 * self.warmup)) / (2 * self.warmup)))
        self.cs = 1 - self.te
        pass



'''
MORTMの学習を行う際にこのモジュールを使います。
train_mortmメソッドを呼び出し、引数の型に合ったオブジェクトを代入してください。
最低でも、「データセット(Tokenizerで変換したもの)のディレクトリ」、「モデルの出力先のディレクトリ」,
「モデルのバージョン」,「ボキャブラリーサイズ」,「エポック回数」、「各トークンの出現回数のリスト」が必要です。
'''

import datetime
import json
import math
import os
import time
from abc import abstractmethod
from typing import  Callable

import torch
from torch import Tensor
from torch.utils.data import DataLoader
import torch.nn as nn
import numpy as np
from torch.optim.lr_scheduler import LambdaLR
from torch.nn.utils.rnn import pad_sequence
from torch.utils.tensorboard import SummaryWriter


from .messager import Messenger, _DefaultMessenger
from .progress import LearningProgress, _DefaultLearningProgress
from .datasets import MORTM_DataSets, MORTMTuringDataset
from .mortm import MORTM
from .noam import noam_lr
from .loss import ReinforceCrossEntropy
from .tokenizer import Tokenizer
IS_DEBUG = False


def _send_prediction_end_time(message, loader_len, begin_time, end_time,
                              vocab_size: int, num_epochs: int, trans_layer, num_heads, d_model,
                              dim_feedforward, dropout, position_length):
    t = end_time - begin_time
    end_time_progress = (t * loader_len * num_epochs) / 3600
    message.send_message("終了見込みについて",
                         f"現在学習が進行しています。\n"
                         f"今回設定したパラメータに基づいて終了時刻を計算しました。\n"
                         f"ボキャブラリーサイズ:{vocab_size}\n"
                         f"エポック回数:{num_epochs}\n"
                         f"Transformerのレイヤー層:{trans_layer}\n"
                         f"Modelの次元数:{d_model}\n"
                         f"シーケンスの長さ:{dim_feedforward}\n"
                         f"ドロップアウト:{dropout}\n"
                         f"\n\n シーケンスの1回目の処理が終了しました。かかった時間は{t:.1f}秒でした。\n"
                         f"終了見込み時間は{end_time_progress:.2f}時間です"
                         )


# デバイスを取得
def _set_train_data(directory, datasets, progress: LearningProgress, is_fine_turing=False):
    print("Starting load....")
    mortm_datasets = MORTM_DataSets(progress) if not is_fine_turing else MORTMTuringDataset(progress)
    loss_count = 0
    count = 0
    dataset_length = 0
    loss_data = 1 if not is_fine_turing else 0
    for dataset in datasets:
        count += 1

        np_load_data = np.load(directory + dataset, allow_pickle=True)
        if len(np_load_data) > loss_data:
            dataset_length += mortm_datasets.add_data(np_load_data)
            print(f"\r {count}/{len(datasets)} | Dataset Length:{dataset_length} | Load[{directory + dataset}]", end="")
        else:
            loss_count += 1
    print("load Successful!!")
    print(f"データセットの規模（曲数）：{len(datasets) - loss_count}")
    print("---------------------------------------")

    return mortm_datasets

def _get_padding_mask(input_ids, progress: LearningProgress):
    # input_ids が Tensor であることを仮定
    pad_id = (input_ids != 0).to(torch.float)
    padding_mask = pad_id.to(progress.get_device())
    return padding_mask


def collate_fn(batch):
    # バッチ内のテンソルの長さを揃える（パディングする）
    batch = pad_sequence(batch, batch_first=True, padding_value=0)
    return batch


def update_log(model, writer, global_step):
    for name, param in model.named_parameters():
        if param.grad is not None:
            writer.add_scalar(f"Gradient Norm/{name}", param.grad.norm(), global_step)

        writer.add_scalar(f"Parameter Value/{name}", param.norm(), global_step)

def get_color(criterion: ReinforceCrossEntropy):
    if criterion.te < 0.1:
        return "\033[32m"
    elif criterion.te < 0.25:
        return "\033[33m"
    elif criterion.te < 0.5:
        return "\033[34m"
    elif criterion.te < 0.75:
        return "\033[35m"
    elif criterion.te < 0.95:
        return "\033[36m"
    else:
        return "\033[37m"

def progress_bar(epoch, sum_epoch, sequence, batch_size, loss, lr, verif_loss, criterion:ReinforceCrossEntropy):
    per = sequence / batch_size * 100
    block = int(per / 100 * 50)
    color_bar = get_color(criterion)
    #color_bar = "\033[32m"
    bar = f" {color_bar}{'#' * block}\033[31m{'-' * (50 - block)}\033[0m"
    print(f"\r learning Epoch {epoch + 1}/{sum_epoch} [{bar}] {per:.2f}%  loss:{loss:.4f} Lr:{lr}  verification loss:{verif_loss: .4f}", end="")


def _train_self_tuning(tokenizer: Tokenizer, save_directory, ayato_dataset, message: Messenger, vocab_size: int, num_epochs: int, weight: Tensor, progress: LearningProgress,
                       trans_layer=6, load_model_directory:str = None, use_rpr=False, src_mask_method: Callable[[Tensor], Tensor]=None,
                       num_heads=8, d_model=512, dim_feedforward=1024, dropout=0.1, is_save_training_progress=False,
                       position_length=2048, accumulation_steps=4, batch_size=16, num_workers=0, warmup_steps=4000, lr_param=1):

    loader = DataLoader(ayato_dataset, batch_size=batch_size, shuffle=True,
                        num_workers=num_workers, collate_fn=collate_fn)

    print("Creating Model....")
    model = MORTM(vocab_size=vocab_size, progress=progress, trans_layer=trans_layer, num_heads=num_heads,
                  d_model=d_model, dim_feedforward=dim_feedforward,
                  dropout=dropout, position_length=position_length, use_rpr=use_rpr).to(progress.get_device())
    if load_model_directory is not None:
        model.load_state_dict(torch.load(load_model_directory))

    #criterion = nn.CrossEntropyLoss(ignore_index=0, weight=weight.to(progress.get_device())).to(progress.get_device())  # 損失関数を定義
    criterion = ReinforceCrossEntropy(tokenizer=tokenizer, ignore_index=0, k=1, warmup=5000, weight=weight.to(progress.get_device()))

    optimizer = torch.optim.Adam(model.parameters(), lr=lr_param, betas=(0.9, 0.98), weight_decay=1e-6)  # オプティマイザを定義
    scheduler = LambdaLR(optimizer=optimizer, lr_lambda=noam_lr(d_model=d_model, warmup_steps=warmup_steps))

    print("Start training...")
    writer = SummaryWriter(save_directory + "/runs/")

    loss_val = None
    mail_bool = True
    for epoch in range(num_epochs):
        try:
            print(f"epoch {epoch + 1} start....")
            count = 1
            epoch_loss = 0.0
            verification_loss = 0.0

            model.train()
            optimizer.zero_grad()

            for inputs in loader:  # seqにはbatch_size分の楽曲が入っている
                #print(f"learning sequence {count}")
                begin_time = time.time()
                input_ids: Tensor = inputs[:, :-1].to(progress.get_device())
                targets: Tensor = inputs[:, 1:].to(progress.get_device())


                padding_mask_in: Tensor = _get_padding_mask(input_ids, progress)

                src_mask = None if src_mask_method is None else src_mask_method(input_ids)

                output = model(inputs_seq=input_ids, src_mask=src_mask, input_padding_mask=padding_mask_in)


                outputs = output.view(-1, output.size(-1)).to(progress.get_device())
                targets = targets.reshape(-1).long()

                loss = criterion(outputs, targets)  # 損失を計算
                epoch_loss += loss.item()
                loss = loss / accumulation_steps
                loss.backward()  # 逆伝播

                update_log(model, writer, count)

                if count % accumulation_steps == 0:  #実質バッチサイズは64である
                    progress.step_optimizer(optimizer, model, accumulation_steps)
                    scheduler.step()
                    #criterion.step()
                    torch.cuda.empty_cache()

                count += 1
                end_time = time.time()

                if mail_bool and message is not None:
                    _send_prediction_end_time(message, len(loader), begin_time, end_time, vocab_size, num_epochs,
                                              trans_layer, num_heads, d_model, dim_feedforward, dropout, position_length)
                    mail_bool = False

                if (count + 1) % message.step_by_message_count == 0:
                    message.send_message("機械学習の途中経過について", f"Epoch {epoch + 1}/{num_epochs}の"
                                                                       f"learning sequence {count}結果は、\n {epoch_loss / count:.4f}でした。\n"
                                                                       f"損失関数スケジューラーは{criterion.cs}です。")
                writer.flush()

                progress_bar(epoch, num_epochs, count, len(loader), epoch_loss / count, scheduler.get_last_lr(), verification_loss, criterion)


            message.send_message("機械学習の途中経過について",
                                     f"Epoch {epoch + 1}/{num_epochs}の結果は、{epoch_loss / count:.4f}でした。"
                                     f"現在の損失関数スケジューラーの重みは{criterion.cs}となっています。")
            loss_val = epoch_loss / count
            writer.add_scalar('Loss/train', epoch_loss / count, epoch)  # 損失値を記録

            if is_save_training_progress:
                torch.save(model.state_dict(), f"{save_directory}/MORTM.train.{epoch}.{epoch_loss / count:.4f}.pth") #エポック終了時に途中経過を保存
                print("途中経過を保存しました。")
        except torch.cuda.OutOfMemoryError or RuntimeError:
            torch.save(model.state_dict(), f"{save_directory}/MORTM.error_end.{epoch}.pth")
            message.send_message("オーバーフローしました・。", f"{epoch}エポック中にオーバーフローが発生しました。\n"
                                                              f"次のエポックに移行します。\n"
                                                              f"現在の損失関数スケジューラーの重みは{criterion.cs}となっています。")
    writer.close()

    return model, loss_val


def _train_fine_turing(save_directory, ayato_dataset, message: Messenger, vocab_size: int, num_epochs: int, weight: Tensor, progress: LearningProgress,
                       trans_layer=6, load_model_directory:str = None, use_rpr=False, begin_tuning_epoch=3, src_mask_method: Callable[[Tensor], Tensor]=None,
                       num_heads=8, d_model=512, dim_feedforward=1024, dropout=0.1, is_save_training_progress=False,
                       position_length=2048, accumulation_steps=4, batch_size=16, num_workers=0, warmup_steps=4000, lr_param=1):

    loader = DataLoader(ayato_dataset, batch_size, shuffle=True, num_workers=num_workers)

    pre_model = MORTM(vocab_size=vocab_size, progress=progress, trans_layer=trans_layer, num_heads=num_heads,
                  d_model=d_model, dim_feedforward=dim_feedforward,
                  dropout=dropout, position_length=position_length, use_rpr=use_rpr).to(progress.get_device())
    pre_model.load_state_dict(torch.load(load_model_directory))

    model = MORTM(vocab_size=vocab_size, progress=progress, trans_layer=trans_layer, num_heads=num_heads,
                      d_model=d_model, dim_feedforward=dim_feedforward,
                      dropout=dropout, position_length=position_length, use_rpr=use_rpr, use_decoder=True).to(progress.get_device())

    model_dict = model.state_dict()

    for param_name, param in pre_model.state_dict().items():
        if param_name in model_dict and "decoder" not in param_name:
            param.requires_grad = False
            model_dict[param_name].copy_(param.data)
        else:
            print(param_name)
    model.load_state_dict(model_dict)



    criterion = nn.CrossEntropyLoss(ignore_index=0, weight=weight.to(progress.get_device())).to(progress.get_device())  # 損失関数を定義


    optimizer = torch.optim.Adam(model.parameters(), lr=lr_param, betas=(0.9, 0.98), weight_decay=1e-6)  # オプティマイザを定義
    scheduler = LambdaLR(optimizer=optimizer, lr_lambda=noam_lr(d_model=d_model, warmup_steps=warmup_steps))

    loss_val = None
    mail_bool = True
    for epoch in range(num_epochs):
        try:
            print(f"epoch {epoch + 1} start....")

            if epoch + 1 == begin_tuning_epoch:
                model_dict = model.state_dict()
                for param_name, param in model_dict.items():
                    param.requires_grad = True
                model.load_state_dict(model_dict)
                print("ファインチューニングをここから開始します。")

            count = 1
            epoch_loss = 0.0
            verification_loss = 0.0

            model.train()
            optimizer.zero_grad()

            for inputs, targets in loader:  # seqにはbatch_size分の楽曲が入っている
                #print(f"learning sequence {count}")
                begin_time = time.time()

                padding_mask_in: Tensor = _get_padding_mask(inputs, progress)
                tgt_mask = model.transformer.generate_square_subsequent_mask(targets.shape[1]).to(progress.get_device())
                #print(padding_mask_in)

                src_mask = None if src_mask_method is None else src_mask_method(inputs)
                output = model(inputs_seq=inputs, src_mask=src_mask, tgt_seq=targets, tgt_mask=tgt_mask, input_padding_mask=padding_mask_in)

                #print(model.softmax(output)[: -1, :][-1, :].argmax())

                outputs = output.view(-1, output.size(-1)).to(progress.get_device())
                targets = targets.reshape(-1).long()
                loss = criterion(outputs, targets)  # 損失を計算
                epoch_loss += loss.item()
                loss = loss / accumulation_steps
                loss.backward()  # 逆伝播

                #update_log(model, writer, count)

                if count % accumulation_steps == 0:  #実質バッチサイズは64である
                    progress.step_optimizer(optimizer, model, accumulation_steps)
                    scheduler.step()

                count += 1
                end_time = time.time()

                if mail_bool and message is not None:
                    _send_prediction_end_time(message, len(loader), begin_time, end_time, vocab_size, num_epochs,
                                              trans_layer, num_heads, d_model, dim_feedforward, dropout, position_length)
                    mail_bool = False

                if (count + 1) % message.step_by_message_count == 0:
                    message.send_message("機械学習の途中経過について", f"Epoch {epoch + 1}/{num_epochs}の"
                                                                       f"learning sequence {count}結果は、\n {epoch_loss / count:.4f}でした。")


                progress_bar(epoch, num_epochs, count, len(loader), epoch_loss / count, scheduler.get_last_lr(), verification_loss)


            message.send_message("機械学習の途中経過について",
                                 f"Epoch {epoch + 1}/{num_epochs}の結果は、{epoch_loss / count:.4f}でした。")
            loss_val = epoch_loss / count

            if is_save_training_progress:
                torch.save(model.state_dict(), f"{save_directory}/MORTM.train.{epoch}.{epoch_loss / count:.4f}.pth") #エポック終了時に途中経過を保存
                print("途中経過を保存しました。")
        except torch.cuda.OutOfMemoryError:
            torch.save(model.state_dict(), f"{save_directory}/MORTM.error_end.{epoch}.pth")

    return model, loss_val




    pass

def train_mortm(tokenizer, dataset_directory, save_directory, version: str, vocab_size: int, num_epochs: int, weight_directory,
                message: Messenger = _DefaultMessenger(), load_model_directory: str=None, use_rpr=True, fine_turing_mode=False,
                trans_layer=9, num_heads=32, d_model=1024, is_save_training_progress=False, lr_param=2e-1, begin_tuning_epoch=3,
                dim_feedforward=4096, dropout=0.2, position_length=8500, num_workers=0, warmup_steps=4000, src_mask_method: Callable[[Tensor], Tensor]=None,
                accumulation_steps=32, batch_size=1, progress: LearningProgress = _DefaultLearningProgress(),):

    os.environ['CUDA_LAUNCH_BLOCKING'] = '1'
    today_date = datetime.date.today().strftime('%Y%m%d')

    print(f"ToDay is{datetime.date.today()}! start generating MORTEM_Model.{version}_{today_date}")

    datasets = os.listdir(dataset_directory)
    train_data = _set_train_data(dataset_directory, datasets, progress, fine_turing_mode)

    try:
        with open(weight_directory, 'r') as file:
            freq_dict = json.load(file)
            # 逆数を取り、頻出度が0の場合は小さい値に設定
            epsilon = 1e-11  # 非ゼロの小さい値を設定しておく
            weights = []

            for i in range(len(freq_dict)):
                freq = freq_dict[str(i)]  # JSONのキーは文字列なのでstrに変換
                if freq == 0:
                    weights.append(epsilon)
                else:
                    weights.append(1.0 / (math.log(freq + 1.0) + epsilon))  # 対数スケーリングを適用
            # テンソルに変換
            weight_tensor = torch.tensor(weights)
            #weight_tensor = weight_tensor / weight_tensor.sum()

            print(weight_tensor[weight_tensor.argmax(dim=-1)], weight_tensor[weight_tensor.argmin(dim=-1)])
        if not fine_turing_mode:
            model, loss = _train_self_tuning(tokenizer,save_directory, train_data, message, vocab_size, num_epochs, weight_tensor, progress=progress,
                                             load_model_directory=load_model_directory,
                                             d_model=d_model,
                                             dim_feedforward=dim_feedforward,
                                             trans_layer=trans_layer,
                                             num_heads=num_heads,
                                             position_length=position_length,
                                             dropout=dropout,
                                             accumulation_steps=accumulation_steps,
                                             batch_size=batch_size,
                                             num_workers=num_workers,
                                             warmup_steps=warmup_steps,
                                             is_save_training_progress=is_save_training_progress,
                                             lr_param=lr_param,
                                             use_rpr=use_rpr,
                                             src_mask_method=src_mask_method
                                             )  # 20エポック分機械学習を行う。
        else:
            model, loss = _train_fine_turing(save_directory, train_data, message, vocab_size, num_epochs, weight_tensor, progress=progress,
                                             load_model_directory=load_model_directory,
                                             d_model=d_model,
                                             dim_feedforward=dim_feedforward,
                                             trans_layer=trans_layer,
                                             num_heads=num_heads,
                                             position_length=position_length,
                                             dropout=dropout,
                                             accumulation_steps=accumulation_steps,
                                             batch_size=batch_size,
                                             num_workers=num_workers,
                                             warmup_steps=warmup_steps,
                                             is_save_training_progress=is_save_training_progress,
                                             lr_param=lr_param,
                                             use_rpr=use_rpr,
                                             begin_tuning_epoch=begin_tuning_epoch,
                                             src_mask_method=src_mask_method
                                             )  # 20エポック分機械学習を行う。

        message.send_message("機械学習終了のお知らせ",
                                 f"MORTM.{version}の機械学習が終了しました。 \n 結果の報告です。\n 損失関数: {loss}")

        torch.save(model.state_dict(), f"{save_directory}/MORTM.{version}_{loss}.pth")  # できたモデルをセーブする

        return model

    except torch.cuda.OutOfMemoryError:
        message.send_message("エラーが発生し、処理を中断しました",
                                 "学習中にモデルがこのPCのメモリーの理論値を超えました。\nバッチサイズを調整してください")
        print("オーバーフローしました。")
    pass



class VerificationLoss:

    def __init__(self, datasets: MORTM_DataSets):
        self.dataloader = DataLoader(dataset=datasets, collate_fn=collate_fn, shuffle=False)

    def __call__(self, *args, **kwargs):
        pass




from torch import Tensor
from pretty_midi import Instrument, Note, PrettyMIDI
from typing import List

from .tokenizer import Tokenizer, PITCH_TYPE, START_TYPE, SHIFT_TYPE, VELOCITY_TYPE, DURATION_TYPE
from .aya_node import Token


def ct_token_to_midi(tokenizer: Tokenizer, seq: Tensor, save_directory:str, program=1):
    seq = seq[1:]
    midi = PrettyMIDI()
    inst: Instrument = Instrument(program=program)
    note = Note(pitch=0, velocity=100, start=0, end=0)
    back_note = None
    token_converter_list = tokenizer.token_list
    for token_id in seq:
        token = tokenizer.rev_get(token_id.item())
        if token_id == 2:
            break
        for con in token_converter_list:
            token_type = con(token=token, back_notes=back_note, note=note)
            if token_type == DURATION_TYPE:
                inst.notes.append(note)
                back_note = note
                note = Note(pitch=0, velocity=100, start=0, end=0)
    midi.instruments.append(inst)

    midi.write(save_directory)
    return midi



import torch
from torch import Tensor
from typing import Callable
import random
from .tokenizer import Tokenizer, START_TYPE, DURATION_TYPE
from .progress import _DefaultLearningProgress
METRIC_RANDOM_MASK = 'mrm'

def get_masks(tokenizer: Tokenizer, get_type:str, progress=_DefaultLearningProgress()) -> Callable[[Tensor], Tensor]:
    def metric_random_mask(src: Tensor) -> Tensor:
        # インデックスの値が200 ~ 400および401 ~ 500である位置を取得
        src = src[-1, :]
        indices_1 = (src >= tokenizer.begin_token(START_TYPE)) & (src <= tokenizer.end_token(START_TYPE))  # 200 ~ 400
        indices_2 = (src >= tokenizer.begin_token(DURATION_TYPE)) & (src <= tokenizer.end_token(DURATION_TYPE))  # 401 ~ 500

        # 20%の確率でインデックスにマスクをかけるためのブール型テンソルを作成
        mask_1 = torch.tensor([random.random() < 0.4 if val else False for val in indices_1])
        mask_2 = torch.tensor([random.random() < 0.4 if val else False for val in indices_2])

        # マスクを結合
        combined_mask = mask_1 | mask_2

        return combined_mask.to(device=progress.get_device())

    if get_type is METRIC_RANDOM_MASK:
        return metric_random_mask


from typing import List, Any

import mido
import numpy as np
from AGSM.convert import ConvTempo
from pretty_midi.pretty_midi import PrettyMIDI, Instrument, Note, TimeSignature
from abc import abstractmethod, ABC
from typing import TypeVar, Generic
from . import constants
from .aya_node import Token
from .tokenizer import Tokenizer

T = TypeVar("T")
class _AbstractMidiToAyaNode(ABC):

    def __init__(self, instance: Generic[T],  tokenizer: Tokenizer, directory: str, file_name: str, program_list, midi_data=None,):
        self.program_list = program_list
        self.directory = directory
        self.file_name = file_name
        self.is_error = False
        self.instance = instance
        self.token_converter: List[Token] = tokenizer.token_list
        self.error_reason: str = "不明なエラー"
        self.tokenizer = tokenizer

        if midi_data is not None:
            self.midi_data: PrettyMIDI = midi_data
        else:
            try:
                self.midi_data: PrettyMIDI = PrettyMIDI(f"{directory}/{file_name}")
                time_s = self.midi_data.time_signature_changes
                for t in time_s:
                    t_s: TimeSignature = t
                    if not (t_s.numerator == 4 and t_s.denominator == 4):
                        self.is_error = True
                        self.error_reason = "旋律に変拍子が混じっていました。"
                        break
            except Exception:
                self.is_error = True
                self.error_reason = "MIDIのロードができませんでした。"

        if not self.is_error:
            self.tempo_change_time, self.tempo = self.midi_data.get_tempo_changes()

    def __call__(self, *args, **kwargs):
        self.convert()

    def get_midi_change_scale(self, scale_up_key):
        midi = PrettyMIDI()

        for ins in self.midi_data.instruments:
            ins: Instrument = ins
            new_inst = Instrument(program=ins.program)
            for note in ins.notes:
                note: Note = note
                pitch = note.pitch + scale_up_key
                if pitch > 127:
                    pitch -= 12
                if pitch < 0:
                    pitch += 12

                start = note.start
                end = note.end
                velo = note.velocity
                new_note = Note(pitch=pitch, velocity=velo, start=start, end=end)
                new_inst.notes.append(new_note)

            midi.instruments.append(new_inst)

        return midi

    def expansion_midi(self) -> List[Any]:
        converts = []
        key = 5
        if not self.is_error:
            for i in range(key):
                midi = self.get_midi_change_scale(i + 1)
                converts.append(self.instance(self.tokenizer, self.directory, f"{self.file_name}_scale_{i + 1}",
                                              self.program_list, midi_data=midi))
                midi = self.get_midi_change_scale(-(i + 1))
                converts.append(self.instance(self.tokenizer, self.directory, f"{self.file_name}_scale_{-(i + 1)}",
                                              self.program_list, midi_data=midi))

        return converts

    def get_tempo(self, start: float):
        tempo = 0
        for i in range(len(self.tempo_change_time)):
            if start >= self.tempo_change_time[i]:
                tempo = self.tempo[i]

        return tempo

    @abstractmethod
    def save(self, save_directory: str) -> [bool, str]:
        pass

    @abstractmethod
    def convert(self):
        pass

class MidiToAyaNode(_AbstractMidiToAyaNode):

    def __init__(self, tokenizer: Tokenizer, directory: str, file_name: str, program_list, midi_data=None):
        super().__init__(MidiToAyaNode, tokenizer, directory, file_name, program_list, midi_data)
        self.aya_node = [0]

    def convert(self):
        """
        以下のステップでMidiが変換される
        1. テンポ固定
            120テンポに変換される。
        2. 指定した楽器があるかを検索する
            インスタンス化した際に、指定したprogram_listに乗っ取り、該当する楽器があるかを検索する
        3. ノートを変換する。
            self.ct_aya_nodeにインストを渡すことで、MORTMに適した形に変換する。

        また、変換したインストゥルメントは以下のような配列構造になる。

        1_music = [
                    [1 inst's 60s clip],
                        ...
                    [1 inst's 60s clip],
                    [2 inst's 60s clip],
                        ...
                    [n inst's 60s clip]
                    ]
        :return:なし。
        """
        if not self.is_error:
            program_count = 0

            for inst in self.midi_data.instruments:
                inst: Instrument = inst
                if not inst.is_drum and inst.program in self.program_list:
                    aya_node_inst = self.ct_aya_node(inst)

                    self.aya_node = self.aya_node + aya_node_inst
                    program_count += 1

            if program_count == 0:
                self.is_error = True
                self.error_reason = f"{self.directory}/{self.file_name}に、欲しい楽器がありませんでした。"

    def ct_aya_node(self, inst: Instrument) -> list:

        """
        Instrumentsから1音ずつ取り出し、Tokenizerで変換する。
        clip = [<START>, S, P, V, D, H, S, P, V, D, H ...<END>]
        さらに60秒ごとにスプリットし、以下のような配列構造を作る。
        aya_node_inst = [clip_1, clip_2 ... clip_n]

        よって、二次元配列のndarrayを返す。
        :param inst: インストゥルメント
        :return: 60秒にクリッピングされた旋律の配列(2次元)
        """

        clip = np.array([], dtype=int)
        aya_node_inst = []
        back_note = None

        clip_time: float = 0
        split_count: int = 1

        sorted_notes = sorted(inst.notes, key=lambda notes: notes.start)

        for note in sorted_notes:
            note: Note = note

            tempo = self.get_tempo(note.start)
            measure_time = (60 / tempo) * 4

            if clip_time >= measure_time * 8 * split_count:
                aya_node_inst = self.marge_clip(clip, aya_node_inst)

                clip = np.array([], dtype=int)
                back_note = None
                split_count += 1

            if back_note is None:
                clip = np.append(clip, self.tokenizer.get(constants.START_SEQ_TOKEN))

            for conv in self.token_converter:
                conv: Token = conv

                token = conv(back_notes=back_note, note=note, tempo=tempo)
                if token is not None:
                    token_id = self.tokenizer.get(token)
                    clip = np.append(clip, token_id)

            back_note = note
            clip_time = note.start

        if len(clip) > 50:
            clip = np.append(clip, self.tokenizer.get(constants.END_SEQ_TOKEN))
            aya_node_inst = self.marge_clip(clip, aya_node_inst)
        return aya_node_inst

    def marge_clip(self, clip, aya_node_inst):
        #clip = np.append(clip, self.tokenizer.get(constants.END_SEQ_TOKEN))
        aya_node_inst.append(clip)

        return aya_node_inst

    def save(self, save_directory: str) -> [bool, str]:
        if not self.is_error:

            array_dict = {f'array{i}': arr for i, arr in enumerate(self.aya_node)}
            if len(array_dict) > 1:
                np.savez(save_directory + "/" + self.file_name, **array_dict)
                return True, "処理が正常に終了しました。"
            else:
                return False, "オブジェクトが何らかの理由で見つかりませんでした。"
        else:
            return False, self.error_reason

    def ct_tempo(self):
        try:
            ct = ConvTempo(directory=self.directory + "/" + self.file_name, midi_data=self.midi_data, change_tempo=120)
            ct.convert()
            self.midi_data = ct.midi_data
        except OSError or IndexError or ValueError or mido.midifiles.meta.KeySignatureError or EOFError or KeyError or ZeroDivisionError:
            self.error_reason = f"{self.directory}/{self.file_name}でテンポ変換中にエラーが発生。処理を中断します。"
            self.is_error = True


class MidiToAyaNode_TGT(_AbstractMidiToAyaNode):

    def save(self, save_directory: str) -> [bool, str]:
        if not self.is_error:

            array_dict = {f'array{i}': arr for i, arr in enumerate(self.aya_node)}
            if len(array_dict) > 1:
                np.savez(save_directory + "/" + self.file_name, **array_dict)
                return True, "処理が正常に終了しました。"
            else:
                return False, "オブジェクトが何らかの理由で見つかりませんでした。"
        else:
            return False, self.error_reason

    def __init__(self, tokenizer: Tokenizer, directory: str, file_name: str, program_list, midi_data=None):
        super().__init__(MidiToAyaNode_TGT, tokenizer, directory, file_name, program_list, midi_data)
        self.aya_node: List[dict] = []

    def convert(self):
        if not self.is_error:
            program_count = 0

            for inst in self.midi_data.instruments:
                inst: Instrument = inst
                if not inst.is_drum and inst.program in self.program_list:
                    aya_node_inst = self.ct_aya_node(inst)

                    self.aya_node = self.aya_node + aya_node_inst
                    program_count += 1

            if program_count == 0:
                self.is_error = True
                self.error_reason = f"{self.directory}/{self.file_name}に、欲しい楽器がありませんでした。"

    def ct_aya_node(self, inst: Instrument) -> list:

        """
            Instrumentsから1音ずつ取り出し、Tokenizerで変換する。
            clip = [<START>, S, P, V, D, H, S, P, V, D, H ...<END>]
            さらに60秒ごとにスプリットし、以下のような配列構造を作る。
            aya_node_inst = [clip_1, clip_2 ... clip_n]

            よって、二次元配列のndarrayを返す。
            :param inst: インストゥルメント
            :return: 60秒にクリッピングされた旋律の配列(2次元)
            """

        clip = np.array([], dtype=int)
        clip_dict: dict = dict()
        aya_node_inst = np.array([], dtype=object)
        is_seq = True
        back_note = None

        clip_time: float = 0
        split_count: int = 1

        sorted_notes = sorted(inst.notes, key=lambda notes: notes.start)

        for note in sorted_notes:
            note: Note = note

            tempo = self.get_tempo(note.start)
            measure_time = (60 / tempo) * 4

            if clip_time >= measure_time * 4 * split_count:
                if is_seq:
                    clip_dict['src'] = clip.tolist()
                else:
                    clip_dict['tgt'] = clip.tolist()
                    aya_node_inst = np.append(aya_node_inst, clip_dict)

                clip_dict = dict()
                clip = np.array([], dtype=int)
                is_seq = not is_seq
                split_count += 1

            if back_note is None:
                clip = np.append(clip, self.tokenizer.get(constants.START_SEQ_TOKEN))

            for conv in self.token_converter:
                conv: Token = conv

                token = conv(back_notes=back_note, note=note, tempo=tempo)
                if token is not None:
                    token_id = self.tokenizer.get(token)
                    clip = np.append(clip, token_id)

            back_note = note
            clip_time = note.start


        return aya_node_inst.tolist()

    def marge_clip(self, clip, aya_node_inst):
        #clip = np.append(clip, self.tokenizer.get(constants.END_SEQ_TOKEN))
        aya_node_inst.append(clip)

        return aya_node_inst
